from matplotlib import cm, pyplot
from matplotlib.axis import Axis
from matplotlib.colors import Normalize


def colorbar_axis(
    normalization: Normalize,
    axis: Axis = None,
    color_map: str = None,
    orientation: str = None,
    own_axis: bool = False,
    extend: str = None,
    label: str = None,
) -> Axis:
    if axis is None:
        figure = pyplot.figure()
        axis = figure.add_subplot(1, 1, 1)

    if color_map is None:
        color_map = 'jet'
    color_map = cm.get_cmap(color_map)

    if orientation is None:
        orientation = 'horizontal'

    if own_axis:
        axis.set_visible(False)
        axis.xaxis.set_visible(False)
        axis.yaxis.set_visible(False)

    return axis.figure.colorbar(
        mappable=cm.ScalarMappable(cmap=color_map, norm=normalization),
        orientation=orientation,
        ax=axis,
        extend=extend,
        label=label,
    )



from os import PathLike
from pathlib import Path
from typing import List

import geodatasets
import geopandas
from matplotlib import pyplot
from matplotlib.cm import get_cmap
from matplotlib.collections import LineCollection
from matplotlib.colors import LogNorm, Normalize
from matplotlib.patches import Patch
import numpy
from stormevents.nhc import VortexTrack
from stormevents.nhc.atcf import ATCF_FileDeck
import xarray

from ensembleperturbation.plotting.surrogate import comparison_plot_grid
from ensembleperturbation.plotting.utilities import colorbar_axis
from ensembleperturbation.utilities import encode_categorical_values


def plot_perturbed_variables(
    perturbations: xarray.Dataset, title: str = None, output_filename: PathLike = None,
):
    figure = pyplot.figure()
    figure.set_size_inches(10, 10 / 1.61803398875)
    if title is None:
        title = f'{len(perturbations["run"])} pertubation(s) of {len(perturbations["variable"])} variable(s)'
    figure.suptitle(title)

    variables = perturbations['variable'].values
    axes, grid = comparison_plot_grid(variables, figure=figure)

    perturbation_colors = perturbations['weights']
    if (
        perturbation_colors.isnull().values.all()
        or (perturbation_colors == perturbation_colors[0]).values.all()
    ):
        perturbation_colors = numpy.arange(len(perturbation_colors))
        perturbation_size = (6 * numpy.ones(len(perturbation_colors))) ** 2
        alpha = 0.9

        normalization = None
        color_map = get_cmap('nipy_spectral')
    else:
        min_value = float(perturbation_colors.min().values)
        max_value = float(perturbation_colors.max().values)

        color_map = get_cmap('Greys')
        orientation = 'horizontal'

        if min_value > 0:
            normalization = LogNorm(vmin=min_value, vmax=max_value)
        else:
            # normalization = SymLogNorm(0.01, vmin=min_value, vmax=max_value)
            normalization = Normalize(vmin=min_value, vmax=max_value)
        colorbar = colorbar_axis(
            normalization=normalization,
            axis=figure.add_subplot(grid[-1, -1]),
            orientation=orientation,
            color_map=color_map,
            own_axis=True,
        )
        colorbar.set_label('weight')

        perturbation_colors.loc[perturbation_colors.isnull()] = 0
        perturbation_size = (
            12 * (perturbation_colors.values - min_value) / (max_value - min_value)
        ) ** 2
        alpha = 0.75

    perturbations = perturbations['perturbations']
    sort_idx = numpy.argsort(perturbation_size)[::-1]
    for row_variable, columns in axes.items():
        for column_variable, axis in columns.items():
            axis.scatter(
                perturbations.sel(variable=column_variable)[sort_idx],
                perturbations.sel(variable=row_variable)[sort_idx],
                c=perturbation_colors[sort_idx],
                s=perturbation_size[sort_idx],
                cmap=color_map,
                norm=normalization,
                alpha=alpha,
                edgecolors=None,
            )

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')


def plot_perturbed_variables_1d(
    perturbations: xarray.Dataset, title: str = None, output_filename: PathLike = None,
):
    figure = pyplot.figure()
    figure.set_size_inches(11, 11 / 1.61803398875)
    if title is None:
        title = f'{len(perturbations["run"])} pertubation(s) of {len(perturbations["variable"])} variable(s)'
    figure.suptitle(title)

    variables = perturbations['variable'].values

    perturbations = perturbations['perturbations']
    for index, variable in enumerate(variables):
        axis = figure.add_subplot(len(variables), 1, index + 1)
        axis.title.set_text(f'{variable}')
        perturbed_var = perturbations.sel(variable=variable)
        axis.scatter(perturbed_var, perturbed_var * 0)
        min_val = perturbed_var.values.min().round(3)
        max_val = perturbed_var.values.max().round(3)
        axis.text(min_val, 0.02, f'min value = {min_val}', ha='left')
        axis.text(max_val, 0.02, f'max value = {max_val}', ha='right')

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')


def plot_perturbations(
    training_perturbations: xarray.Dataset,
    validation_perturbations: xarray.Dataset,
    runs: List[str],
    perturbation_types: List[str],
    track_directory: PathLike = None,
    output_directory: PathLike = None,
):
    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)

    plot_perturbed_variables(
        training_perturbations,
        title=f'{len(training_perturbations["run"])} training pertubation(s) of {len(training_perturbations["variable"])} variable(s)',
        output_filename=output_directory / 'training_perturbations.png'
        if output_directory is not None
        else None,
    )
    plot_perturbed_variables_1d(
        training_perturbations,
        title=f'{len(training_perturbations["run"])} training pertubation(s) of {len(training_perturbations["variable"])} variable(s)',
        output_filename=output_directory / 'training_perturbations_1d.png'
        if output_directory is not None
        else None,
    )

    plot_perturbed_variables(
        validation_perturbations,
        title=f'{len(validation_perturbations["run"])} validation pertubation(s) of {len(validation_perturbations["variable"])} variable(s)',
        output_filename=output_directory / 'validation_perturbations.png'
        if output_directory is not None
        else None,
    )
    plot_perturbed_variables_1d(
        validation_perturbations,
        title=f'{len(validation_perturbations["run"])} validation pertubation(s) of {len(validation_perturbations["variable"])} variable(s)',
        output_filename=output_directory / 'validation_perturbations_1d.png'
        if output_directory is not None
        else None,
    )

    if track_directory is not None:
        if not isinstance(track_directory, Path):
            track_directory = Path(track_directory)

        if track_directory.exists():
            track_filenames = {
                track_filename.stem: track_filename
                for track_filename in track_directory.glob('*.22')
            }

            num_perturbations = len(runs)
            if 'original' in track_filenames.keys():
                runs = numpy.append(runs, 'original')
                perturbation_types = numpy.append(perturbation_types, 'original')

            figure = pyplot.figure()
            figure.set_size_inches(12, 12 / 1.61803398875)
            figure.suptitle(f'{num_perturbations} perturbations of storm track')

            map_axis = figure.add_subplot(1, 1, 1)
            countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))

            unique_perturbation_types = numpy.unique(perturbation_types)
            encoded_perturbation_types = encode_categorical_values(
                perturbation_types, unique_values=unique_perturbation_types
            )
            linear_normalization = Normalize()
            colors = get_cmap('jet')(linear_normalization(encoded_perturbation_types))

            bounds = numpy.array([None, None, None, None])
            for index, run in enumerate(runs):
                storm = VortexTrack.from_file(track_filenames[run]).data
                points = storm.loc[:, ['longitude', 'latitude']].values.reshape(-1, 1, 2)
                segments = numpy.concatenate([points[:-1], points[1:]], axis=1)
                line_collection = LineCollection(
                    segments,
                    linewidths=numpy.concatenate(
                        [
                            [0],
                            storm['radius_of_maximum_winds']
                            / max(storm['radius_of_maximum_winds']),
                        ]
                    )
                    * 4,
                    color=colors[index],
                )
                map_axis.add_collection(line_collection)

                track_bounds = numpy.array(
                    [
                        points[:, :, 0].min(),
                        points[:, :, 1].min(),
                        points[:, :, 0].max(),
                        points[:, :, 1].max(),
                    ]
                )
                if bounds[0] is None or track_bounds[0] < bounds[0]:
                    bounds[0] = track_bounds[0]
                if bounds[1] is None or track_bounds[1] < bounds[1]:
                    bounds[1] = track_bounds[1]
                if bounds[2] is None or track_bounds[2] > bounds[2]:
                    bounds[2] = track_bounds[2]
                if bounds[3] is None or track_bounds[3] > bounds[3]:
                    bounds[3] = track_bounds[3]

            map_axis.set_xlim((bounds[0], bounds[2]))
            map_axis.set_ylim((bounds[1], bounds[3]))

            unique_perturbation_type_colors = get_cmap('jet')(
                linear_normalization(numpy.unique(encoded_perturbation_types))
            )
            map_axis.legend(
                [Patch(facecolor=color) for color in unique_perturbation_type_colors],
                unique_perturbation_types,
            )

            xlim = map_axis.get_xlim()
            ylim = map_axis.get_ylim()

            countries.plot(color='lightgrey', ax=map_axis)

            map_axis.set_xlim(xlim)
            map_axis.set_ylim(ylim)

            if output_directory is not None:
                figure.savefig(
                    output_directory / 'storm_tracks.png', dpi=200, bbox_inches='tight',
                )


def plot_track_perturbations(
    perturbations: dict,
    storm_name: str = None,
    output_directory: PathLike = None,
    file_deck: ATCF_FileDeck = None,
    advisories: List[str] = None,
):

    num_perturbations = len(perturbations)
    perturbation_keys = list(perturbations.keys())
    perturbation_types = ['ensemble'] * num_perturbations
    if 'original' in perturbation_keys:
        num_perturbations = num_perturbations - 1
        perturbation_keys.remove('original')
        perturbation_keys.append('original')
        perturbation_types[-1] = 'original'

    figure = pyplot.figure()
    figure.set_size_inches(12, 12)
    figure.suptitle(f'{num_perturbations} perturbations of {storm_name}')

    map_axis = figure.add_subplot(2, 1, 1)
    Vmax_axis = figure.add_subplot(2, 2, 3)
    Rmax_axis = figure.add_subplot(2, 2, 4)
    countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))

    unique_perturbation_types = numpy.unique(perturbation_types)
    encoded_perturbation_types = encode_categorical_values(
        perturbation_types, unique_values=unique_perturbation_types
    )
    linear_normalization = Normalize()
    colors = get_cmap('jet')(linear_normalization(encoded_perturbation_types))

    bounds = numpy.array([None, None, None, None])
    for index, run in enumerate(perturbation_keys):
        pkey = list(perturbations[run].keys())
        storm = VortexTrack.from_file(
            perturbations[run][pkey[0]]['fort22_filename'],
            file_deck=file_deck,
            advisories=advisories,
        ).data

        # Track
        points = storm.loc[:, ['longitude', 'latitude']].values.reshape(-1, 1, 2)
        segments = numpy.concatenate([points[:-1], points[1:]], axis=1)
        line_collection = LineCollection(
            segments,
            linewidths=numpy.concatenate(
                [[0], storm['radius_of_maximum_winds'] / 200,]  # maximum Rmax
            )
            * 4,
            color=colors[index],
        )
        map_axis.add_collection(line_collection)

        track_bounds = numpy.array(
            [
                points[:, :, 0].min(),
                points[:, :, 1].min(),
                points[:, :, 0].max(),
                points[:, :, 1].max(),
            ]
        )
        if bounds[0] is None or track_bounds[0] < bounds[0]:
            bounds[0] = track_bounds[0]
        if bounds[1] is None or track_bounds[1] < bounds[1]:
            bounds[1] = track_bounds[1]
        if bounds[2] is None or track_bounds[2] > bounds[2]:
            bounds[2] = track_bounds[2]
        if bounds[3] is None or track_bounds[3] > bounds[3]:
            bounds[3] = track_bounds[3]

        map_axis.set_xlim((bounds[0], bounds[2]))
        map_axis.set_ylim((bounds[1], bounds[3]))

        # Vmax
        Vmax_axis.plot(
            storm['datetime'],
            storm['max_sustained_wind_speed'],
            color=colors[index],
            linewidth=0.75 if run == 'original' else 0.25,
        )

        # Rmax
        Rmax_axis.plot(
            storm['datetime'],
            storm['radius_of_maximum_winds'],
            color=colors[index],
            linewidth=0.75 if run == 'original' else 0.25,
        )

    unique_perturbation_type_colors = get_cmap('jet')(
        linear_normalization(numpy.unique(encoded_perturbation_types))
    )
    map_axis.legend(
        [Patch(facecolor=color) for color in unique_perturbation_type_colors],
        unique_perturbation_types,
    )

    xlim = map_axis.get_xlim()
    ylim = map_axis.get_ylim()

    countries.plot(color='lightgrey', ax=map_axis)

    map_axis.set_xlim(xlim)
    map_axis.set_ylim(ylim)

    Rmax_axis.grid()
    Vmax_axis.grid()

    Rmax_axis.set_ylabel('$R_{max}$ [n mi]')
    Vmax_axis.set_ylabel('$V_{max}$ [kt]')

    Rmax_axis.set_xticklabels(Rmax_axis.get_xticklabels(), rotation=45)
    Vmax_axis.set_xticklabels(Vmax_axis.get_xticklabels(), rotation=45)

    Rmax_axis.set_title('Size')
    Vmax_axis.set_title('Intensity')
    map_axis.set_title('Trajectory')

    if output_directory is not None:
        figure.savefig(
            output_directory / f'{storm_name}_perturbations.png', dpi=200, bbox_inches='tight',
        )



import io
from os import PathLike
import pathlib
import zipfile

import appdirs
import geopandas
from matplotlib import pyplot
from matplotlib.axis import Axis
import requests


def download_coastline(overwrite: bool = False) -> pathlib.Path:
    data_directory = pathlib.Path(appdirs.user_data_dir('ne_coastline'))
    if not data_directory.exists():
        data_directory.mkdir(exist_ok=True, parents=True)

    coastline_filename = data_directory / 'ne_110m_coastline.shp'

    if not coastline_filename.exists() or overwrite:
        # download and save if not present
        url = 'http://naciscdn.org/naturalearth/110m/physical/ne_110m_coastline.zip'
        response = requests.get(url, stream=True)
        with zipfile.ZipFile(io.BytesIO(response.content)) as zip_file:
            for member_filename in zip_file.namelist():
                file_data = zip_file.read(member_filename)
                with open(data_directory / member_filename, 'wb') as output_file:
                    output_file.write(file_data)
        assert coastline_filename.exists(), 'coastline file not downloaded'

    return coastline_filename


def plot_coastline(
    axis: Axis = None, show: bool = False, save_filename: PathLike = None, **kwargs
):
    if axis is None:
        figure = pyplot.figure()
        axis = figure.add_subplot(1, 1, 1)

    coastline_filename = download_coastline()
    dataframe = geopandas.read_file(coastline_filename)
    dataframe.plot(ax=axis, **kwargs)

    if save_filename is not None:
        pyplot.savefig(save_filename)

    if show:
        pyplot.show()



from concurrent.futures import ProcessPoolExecutor
import logging
from os import PathLike
from pathlib import Path
import sys
import traceback

import numpy
import pint
from pint_pandas import PintType
from pyproj import CRS, Geod, Transformer
from shapely.geometry import Point

units = pint.UnitRegistry()
PintType.ureg = units


def repository_root(path: PathLike = None) -> Path:
    if path is None:
        path = __file__
    if not isinstance(path, Path):
        path = Path(path)
    if path.is_file():
        path = path.parent
    if '.git' in (child.name for child in path.iterdir()) or path == path.parent:
        return path
    else:
        return repository_root(path.parent)


def get_logger(
    name: str,
    log_filename: PathLike = None,
    file_level: int = None,
    console_level: int = None,
    log_format: str = None,
) -> logging.Logger:
    """
    instantiate logger instance

    :param name: name of logger
    :param log_filename: path to log file
    :param file_level: minimum log level to write to log file
    :param console_level: minimum log level to print to console
    :param log_format: logger message format
    :return: instance of a Logger object
    """

    if file_level is None:
        file_level = logging.DEBUG
    if console_level is None:
        console_level = logging.INFO
    logger = logging.getLogger(name)

    # check if logger is already configured
    if logger.level == logging.NOTSET and len(logger.handlers) == 0:
        # check if logger has a parent
        if '.' in name:
            if isinstance(logger.parent, logging.RootLogger):
                for existing_console_handler in [
                    handler
                    for handler in logger.parent.handlers
                    if not isinstance(handler, logging.FileHandler)
                ]:
                    logger.parent.removeHandler(existing_console_handler)
            logger.parent = get_logger(name.rsplit('.', 1)[0])
        else:
            # otherwise create a new split-console logger
            if console_level != logging.NOTSET:
                for existing_console_handler in [
                    handler
                    for handler in logger.handlers
                    if not isinstance(handler, logging.FileHandler)
                ]:
                    logger.removeHandler(existing_console_handler)

                console_output = logging.StreamHandler(sys.stdout)
                console_output.setLevel(console_level)
                logger.addHandler(console_output)

    if log_filename is not None:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(file_level)
        for existing_file_handler in [
            handler for handler in logger.handlers if isinstance(handler, logging.FileHandler)
        ]:
            logger.removeHandler(existing_file_handler)
        logger.addHandler(file_handler)

    if log_format is None:
        log_format = '[%(asctime)s] %(name)-15s %(levelname)-8s: %(message)s'
    log_formatter = logging.Formatter(log_format)
    for handler in logger.handlers:
        handler.setFormatter(log_formatter)

    return logger


def ellipsoidal_distance(
    point_a: (float, float), point_b: (float, float), crs_a: CRS, crs_b: CRS = None
) -> float:
    if isinstance(point_a, Point):
        point_a = [*point_a.coords]
    if isinstance(point_b, Point):
        point_b = [*point_b.coords]
    if crs_b is not None:
        transformer = Transformer.from_crs(crs_b, crs_a)
        point_b = transformer.transform(*point_b)
    points = numpy.stack((point_a, point_b), axis=0)
    ellipsoid = crs_a.datum.to_json_dict()['ellipsoid']
    geodetic = Geod(a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening'])
    return geodetic.line_length(points[:, 0], points[:, 1])


class ProcessPoolExecutorStackTraced(ProcessPoolExecutor):
    def submit(self, fn, *args, **kwargs):
        """Submits the wrapped function instead of `fn`"""

        return super(ProcessPoolExecutorStackTraced, self).submit(
            self._function_wrapper, fn, *args, **kwargs,
        )

    @staticmethod
    def _function_wrapper(fn, *args, **kwargs):
        """
        Wraps `fn` in order to preserve the traceback of any kind of raised exception
        """

        try:
            return fn(*args, **kwargs)
        except Exception:
            # Creates an exception of the same type with the traceback as message
            raise sys.exc_info()[0](traceback.format_exc())


def encode_categorical_values(values: list, unique_values: list = None) -> list:
    if unique_values is None:
        unique_values = numpy.unique(values)

    if not isinstance(values, numpy.ndarray) or len(values.shape) == 1:
        return numpy.concatenate(
            [numpy.where(unique_values == value) for value in values]
        ).squeeze()
    else:
        rows = []
        for row in values:
            rows.append(encode_categorical_values(row, unique_values=unique_values))
        return numpy.stack(rows, axis=0)


def move_to_end(lst, elem):
    # moving elem to end of list
    lst = [x for x in lst if not isinstance(x, elem)] + [x for x in lst if isinstance(x, elem)]
    return lst



from argparse import ArgumentParser
import logging
from os import PathLike
from pathlib import Path
from typing import Dict, List, Tuple, Union

from pandas import DataFrame

from ensembleperturbation.parsing.adcirc import combine_outputs as combine_outputs_adcirc
from ensembleperturbation.parsing.adcirc import ElevationSelection as AdcircElevationSelection
from ensembleperturbation.parsing.schism import combine_outputs as combine_outputs_schism
from ensembleperturbation.parsing.schism import convert_schism_output_files_to_adcirc_like
from ensembleperturbation.parsing.schism import ElevationSelection as SchismElevationSelection
from ensembleperturbation.utilities import get_logger

LOGGER = get_logger('parsing')


def parse_combine_results():
    cwd = Path.cwd()

    argument_parser = ArgumentParser()

    argument_parser.add_argument('--adcirc', action='store_true')
    argument_parser.add_argument('--schism', action='store_true')
    argument_parser.add_argument('--adcirc-like-output', action='store_true')
    argument_parser.add_argument('output', nargs='?', default=cwd, help='output directory')
    argument_parser.add_argument(
        'directory',
        nargs='?',
        default=cwd,
        help='directory containing completed `runs` directory',
    )
    argument_parser.add_argument(
        '--filenames', nargs='*', default=None, help='Model output files to parse',
    )
    argument_parser.add_argument(
        '--bounds', help='bounding box in format `(minx,miny,maxx,maxy)`'
    )
    argument_parser.add_argument('--max-depth', help='maximum depth value to filter by')
    argument_parser.add_argument(
        '--elevation-selection',
        help='filter elevation nodes based on sea level (one of `wet`, `inundated`, `dry`)',
    )
    argument_parser.add_argument(
        '--parallel', action='store_true', help='load concurrently with Dask'
    )
    argument_parser.add_argument(
        '--verbose', action='store_true', help='log more verbose messages'
    )
    arguments = argument_parser.parse_args()

    return {
        'model': 'schism' if arguments.schism else 'adcirc',
        'output': arguments.output,
        'adcirc_like': arguments.adcirc_like_output,
        'directory': arguments.directory,
        'filenames': arguments.filenames,
        'max_depth': float(arguments.max_depth) if arguments.max_depth is not None else None,
        'bounds': arguments.bounds,
        'elevation_selection': arguments.elevation_selection,
        'parallel': arguments.parallel,
        'verbose': arguments.verbose,
    }


def combine_results(
    output: PathLike,
    model: str = 'adcirc',
    adcirc_like: bool = False,
    directory: PathLike = None,
    filenames: List[str] = None,
    bounds: Tuple[float, float, float, float] = None,
    max_depth: float = None,
    elevation_selection: Union[AdcircElevationSelection, SchismElevationSelection] = None,
    parallel: bool = False,
    verbose: bool = False,
) -> Dict[str, DataFrame]:
    if verbose:
        get_logger(LOGGER.name, console_level=logging.DEBUG)

    combine_func = None
    if model == 'adcirc':
        combine_func = combine_outputs_adcirc
    elif model == 'schism':
        combine_func = combine_outputs_schism
        if adcirc_like:
            combine_func = convert_schism_output_files_to_adcirc_like
    else:
        raise ValueError(f'Invalid model specific: "{model}"')

    parsed_data = combine_func(
        directory,
        file_data_variables=filenames,
        bounds=bounds,
        maximum_depth=max_depth,
        elevation_selection=elevation_selection,
        output_directory=output,
        parallel=parallel,
    )

    return parsed_data


def main():
    combine_results(**parse_combine_results())


if __name__ == '__main__':
    main()



from pathlib import Path
from argparse import ArgumentParser

from coupledmodeldriver.client.initialize_adcirc import (
    initialize_adcirc,
    parse_initialize_adcirc_arguments,
)
from coupledmodeldriver.client.initialize_schism import (
    initialize_schism,
    parse_initialize_schism_arguments,
)
from coupledmodeldriver.configure import BestTrackForcingJSON

from ensembleperturbation.perturbation.atcf import perturb_tracks


def main():

    # Using a preliminary parser to decide whether to create
    # "adcirc" or "schism" parser below (args of this parser need
    # to be passed as extra_arguments to the subsequent parser)
    preliminary_parser = ArgumentParser(add_help=False)
    preliminary_parser.add_argument(
        '--adcirc', action='store_true',
    )
    preliminary_parser.add_argument(
        '--schism', action='store_true',
    )
    prelim_args, _ = preliminary_parser.parse_known_args()

    init_schism = prelim_args.schism
    init_adcirc = prelim_args.adcirc or not init_schism

    if init_adcirc:
        parse_func = parse_initialize_adcirc_arguments
        init_func = initialize_adcirc
    elif init_schism:
        parse_func = parse_initialize_schism_arguments
        init_func = initialize_schism

    arguments = parse_func(
        extra_arguments={
            'adcirc': (bool, 'Initialize ADCIRC simulation ensemble (default)'),
            'schism': (bool, 'Initialize SCHISM simulation ensemble'),
            'perturbations': (int, 'number of perturbations to create'),
            'variables': ([str], 'vortex variables to perturb'),
            'sample': (
                bool,
                'override given perturbations with random samples from the joint distribution',
            ),
            'sample-rule': (
                str,
                "rule to use for the distribution sampling. Please choose from: 'random' [default], 'sobol', 'halton', 'hammersley', 'korobov', 'additive_recursion', or 'latin_hypercube'",
            ),
            'quadrature': (bool, 'add additional perturbations along the quadrature'),
        }
    )

    ocean_model_specific_args = {}
    if init_adcirc:
        ocean_model_specific_args = dict(
            adcirc_executable=arguments['adcirc_executable'],
            adcprep_executable=arguments['adcprep_executable'],
            aswip_executable=arguments['aswip_executable'],
            adcirc_processors=arguments['adcirc_processors'],
        )

    elif init_schism:
        ocean_model_specific_args = dict(
            schism_executable=arguments['schism_executable'],
            schism_processors=arguments['schism_processors'],
            schism_hotstart_combiner=arguments['schism_hotstart_combiner'],
            schism_schout_combiner=arguments['schism_schout_combiner'],
            schism_use_old_io=arguments['schism_use_old_io'],
        )

    if arguments['perturbations'] is None:
        raise ValueError('number of perturbations not given')

    if arguments['variables'] is None or len(arguments['variables']) == 0:
        arguments['variables'] = [
            'cross_track',
            'along_track',
            'radius_of_maximum_winds',
            'max_sustained_wind_speed',
        ]

    track_directory = arguments['output_directory'] / 'track_files'
    if not track_directory.exists():
        track_directory.mkdir(parents=True, exist_ok=True)
    original_track_filename = track_directory / 'original.22'

    if original_track_filename.exists():
        storm = original_track_filename

        vortex_forcing = BestTrackForcingJSON.from_fort22(
            original_track_filename,
            start_date=arguments['modeled_start_time'],
            end_date=arguments['modeled_start_time'] + arguments['modeled_duration'],
        )
        arguments['forcings'].append(vortex_forcing)
        for index, forcing in enumerate(arguments['forcings']):
            if isinstance(forcing, BestTrackForcingJSON):
                arguments['forcings'][index] = vortex_forcing
                break
    else:
        for forcing in arguments['forcings']:
            if isinstance(forcing, BestTrackForcingJSON):
                storm = forcing.pyschism_forcing.nhc_code
                break
        else:
            raise ValueError('no best track forcing specified')

    perturbations = perturb_tracks(
        perturbations=arguments['perturbations'],
        directory=Path(arguments['output_directory']) / 'track_files',
        storm=storm,
        variables=arguments['variables'],
        sample_from_distribution=arguments['sample'],
        sample_rule=arguments['sample-rule'],
        quadrature=arguments['quadrature'],
        start_date=arguments['modeled_start_time'],
        end_date=arguments['modeled_start_time'] + arguments['modeled_duration'],
        overwrite=arguments['overwrite'],
    )

    init_func(
        platform=arguments['platform'],
        mesh_directory=arguments['mesh_directory'],
        modeled_start_time=arguments['modeled_start_time'],
        modeled_duration=arguments['modeled_duration'],
        modeled_timestep=arguments['modeled_timestep'],
        tidal_spinup_duration=arguments['tidal_spinup_duration'],
        perturbations=perturbations,
        nems_connections=None,
        nems_mediations=None,
        nems_sequence=None,
        nems_interval=arguments['nems_interval'],
        modulefile=arguments['modulefile'],
        forcings=arguments['forcings'],
        job_duration=arguments['job_duration'],
        output_directory=arguments['output_directory'],
        absolute_paths=arguments['absolute_paths'],
        overwrite=arguments['overwrite'],
        verbose=arguments['verbose'],
        **ocean_model_specific_args,
    )


if __name__ == '__main__':
    main()



from argparse import ArgumentParser

import cartopy.crs as ccrs
from cartopy.mpl.gridliner import LATITUDE_FORMATTER, LONGITUDE_FORMATTER
import cmocean
from matplotlib import pyplot
import pandas
from pandas import DataFrame


def main():
    argument_parser = ArgumentParser()
    argument_parser.add_argument(
        'filename', help='filename (`*.h5`) generated by `combine_results`'
    )
    arguments = argument_parser.parse_args()

    combined_results_filename = arguments.filename

    input_dataframe: DataFrame = pandas.read_hdf(
        combined_results_filename, key='vortex_perturbation_parameters'
    )
    output_dataframe: DataFrame = pandas.read_hdf(combined_results_filename, key='zeta_max')

    # just get standard deviation of the outputs
    output_standard_deviation = output_dataframe.std(axis=1, skipna=True)

    # Plot variable distributions the histogram of the data
    for variable in input_dataframe:
        figure = pyplot.figure()
        axis = figure.add_subplot(1, 1, 1)
        axis.hist(input_dataframe[variable], bins='auto', density=True, alpha=0.75)
        axis.set_xlabel('Parameter')
        axis.set_ylabel('Probability')
        axis.set_title('Histogram of ' + variable)
        axis.grid(True)
        pyplot.show()

    # plot histogram
    figure = pyplot.figure()
    axis = figure.add_subplot(1, 1, 1)
    axis.hist(output_standard_deviation, bins='auto', density=True, alpha=0.75)
    axis.set_xlabel('Parameter')
    axis.set_ylabel('Probability')
    axis.set_title('Histogram of Maximum Elevation Variability')
    axis.grid(True)
    pyplot.show()

    # plot map
    figure = pyplot.figure(figsize=(18, 8))
    axis = figure.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
    axis.scatter(
        output_dataframe.x,
        output_dataframe.y,
        s=1,
        c=output_standard_deviation,
        transform=ccrs.PlateCarree(),
        cmap=cmocean.cm.amp,
        vmin=0,
        vmax=1.6,
    )
    axis.coastlines()

    # pyplot.colorbar(ax=axis, shrink=0.98, extend='max', label='STD [m]')

    # Add the gridlines
    gridlines = axis.gridlines(color='black', linestyle='dotted', draw_labels=True, alpha=0.5)
    gridlines.top_labels = None
    gridlines.right_labels = None
    gridlines.xformatter = LONGITUDE_FORMATTER
    gridlines.yformatter = LATITUDE_FORMATTER

    figure.suptitle('Maximum Elevation Variability from 40-member ensemble')
    pyplot.show()


if __name__ == '__main__':
    main()



from argparse import ArgumentParser

from stormevents.nhc.atcf import ATCF_FileDeck
import typepigeon

from ensembleperturbation.perturbation.atcf import (
    AlongTrack,
    CrossTrack,
    MaximumSustainedWindSpeed,
    perturb_tracks,
    RadiusOfMaximumWinds,
)


def main():
    ##################################
    # Example calls from command line for 2018 Hurricane Florence:
    # - python3 make_storm_ensemble.py 3 al062018 2018-09-11-06 2018-09-17-06
    # - python3 make_storm_ensemble.py 5 Florence2018 2018-09-11-06
    ##################################

    # Implement argument parsing
    argument_parser = ArgumentParser()
    argument_parser.add_argument(
        '--number-of-perturbations', required=True, type=int, help='number of perturbations'
    )
    argument_parser.add_argument('--storm-code', required=True, help='storm name/code')
    argument_parser.add_argument('--start-date', help='start date')
    argument_parser.add_argument('--end-date', help='end date')
    argument_parser.add_argument(
        '--advisories', nargs='*', help='record type (i.e. `BEST`, `OFCL`)'
    )
    argument_parser.add_argument(
        '--file-deck', help='letter of file deck, one of `a`, `b`, or `f`'
    )
    argument_parser.add_argument('directory', nargs='?', help='output directory')
    arguments = argument_parser.parse_args()

    # hardcoding variable list for now
    variables = [
        MaximumSustainedWindSpeed,
        RadiusOfMaximumWinds,
        AlongTrack,
        CrossTrack,
    ]

    perturb_tracks(
        perturbations=arguments.number_of_perturbations,
        directory=arguments.directory,
        storm=arguments.storm_code,
        variables=variables,
        start_date=arguments.start_date,
        end_date=arguments.end_date,
        file_deck=typepigeon.convert_value(arguments.file_deck, ATCF_FileDeck),
        advisories=arguments.advisories,
    )


if __name__ == '__main__':
    main()



from os import PathLike

import cartopy
import geodatasets
import geopandas
from matplotlib import cm, gridspec, pyplot
from matplotlib.axis import Axis
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.tri import Triangulation
import numpy
from stormevents.nhc import VortexTrack
import xarray

from ensembleperturbation.plotting.utilities import colorbar_axis


def node_color_map(
    nodes: xarray.Dataset,
    colors: list = None,
    color_map: str = None,
    min_value: float = None,
    max_value: float = None,
    logarithmic: bool = False,
) -> (numpy.ndarray, Normalize, Colormap, numpy.ndarray):
    if color_map is None:
        color_map = cm.get_cmap('plasma')
    else:
        color_map = cm.get_cmap(color_map)

    if colors is None:
        color_values = numpy.arange(len(nodes['node']))

        normalization = Normalize(vmin=numpy.min(color_values), vmax=numpy.max(color_values))
        colors = color_map(normalization(color_values))
    elif isinstance(colors, str):
        color_values = nodes[colors]

        if len(color_values.dims) > 1:
            color_values = color_values.mean(
                [dim for dim in color_values.dims if dim != 'node']
            )
        if min_value is None:
            min_value = float(color_values.min().values)
        if max_value is None:
            max_value = float(color_values.max().values)
        try:
            normalization = LogNorm(vmin=min_value, vmax=max_value)
            normalized_color_values = normalization(color_values)
        except ValueError:
            normalization = Normalize(vmin=min_value, vmax=max_value)
            normalized_color_values = normalization(color_values)
        colors = color_map(normalized_color_values)
    else:
        colors = numpy.array(colors)

        if min_value is None:
            min_value = numpy.nanmin(colors)
        if max_value is None:
            max_value = numpy.nanmax(colors)

        if logarithmic:
            normalization = LogNorm(vmin=min_value, vmax=max_value)
        else:
            normalization = Normalize(vmin=min_value, vmax=max_value)

        if (len(colors.shape) < 2 or colors.shape[1] != 4) and numpy.any(~numpy.isnan(colors)):
            color_values = colors
            colors = color_map(normalization(color_values))
        else:
            color_values = None

    return color_values, normalization, color_map, colors


def plot_nodes_across_runs(
    nodes: xarray.Dataset,
    title: str = None,
    colors: [] = None,
    storm: str = None,
    output_filename: PathLike = None,
    min_value: float = None,
    max_value: float = None,
    logarithmic: bool = False,
):
    figure = pyplot.figure()
    figure.set_size_inches(12, 12 / 1.61803398875)
    if title is not None:
        figure.suptitle(title)

    grid = gridspec.GridSpec(len(nodes.data_vars), 2, figure=figure)

    map_crs = cartopy.crs.PlateCarree()
    map_axis = figure.add_subplot(grid[:, 0], projection=map_crs)

    color_values, normalization, color_map, map_colors = node_color_map(
        nodes, colors=colors, min_value=min_value, max_value=max_value, logarithmic=logarithmic
    )
    plot_node_map(
        nodes, colors=map_colors, storm=storm, map_axis=map_axis, logarithmic=logarithmic
    )

    if colors is not None:
        colorbar_axis(
            normalization=Normalize(
                vmin=numpy.nanmin(color_values), vmax=numpy.nanmax(color_values)
            ),
            axis=map_axis,
            orientation='vertical',
        )

    shared_axis = None
    for variable_index, (variable_name, variable) in enumerate(nodes.data_vars.items()):
        axis_kwargs = {}
        if shared_axis is not None:
            axis_kwargs['sharex'] = shared_axis

        variable_axis = figure.add_subplot(grid[variable_index, 1], **axis_kwargs)

        if shared_axis is None:
            shared_axis = variable_axis

        if variable_index < len(nodes.data_vars) - 1:
            variable_axis.get_xaxis().set_visible(False)

        if 'source' in variable.dims:
            sources = variable['source']
        else:
            sources = [None]

        for source_index, source in enumerate(sources):
            kwargs = {}
            if source == 'model':
                variable_colors = cm.get_cmap('jet')(color_values)
            elif source == 'surrogate':
                variable_colors = 'grey'
                kwargs['linestyle'] = '--'
            else:
                variable_colors = cm.get_cmap('jet')(color_values)

            if 'source' in variable.dims:
                source_data = variable.sel(source=source)
            else:
                source_data = variable

            if 'time' in nodes.dims:
                for node_index in range(len(nodes['node'])):
                    node_data = source_data.isel(node=node_index)
                    node_color = variable_colors[node_index]
                    node_data.plot.line(
                        x='time', c=node_color, ax=variable_axis, **kwargs,
                    )
                    if variable_name == 'mean' and 'std' in nodes.data_vars:
                        std_data = nodes['std'].isel(node=node_index)
                        if 'source' in std_data.dims:
                            std_data = std_data.sel(source=source)
                        variable_axis.fill_between(
                            nodes['time'],
                            node_data - std_data,
                            node_data + std_data,
                            color=node_color,
                            alpha=0.3,
                            **kwargs,
                        )
            else:
                bar_width = 1
                bar_offset = bar_width * (source_index + 0.5 - len(sources) / 2)

                variable_axis.bar(
                    x=numpy.arange(len(source_data)) + bar_offset,
                    width=bar_width,
                    height=source_data,
                    color=variable_colors,
                    **kwargs,
                )
                variable_axis.set_ylim([min(0, source_data.min()), source_data.max()])

        variable_axis.set_title(variable_name)
        variable_axis.tick_params(axis='x', which='both', labelsize=6)
        variable_axis.set(xlabel=None)

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')


def plot_node_map(
    nodes: xarray.Dataset,
    map_title: str = None,
    colors: list = None,
    color_map: str = None,
    storm: str = None,
    map_axis: Axis = None,
    min_value: float = None,
    max_value: float = None,
    logarithmic: bool = False,
    **kwargs,
):
    if isinstance(colors, str) and map_title is not None:
        map_title = f'"{colors}" of {map_title}'

    data_var_name = [i for i in nodes.data_vars]
    color_values, normalization, color_map, colors = node_color_map(
        nodes,
        colors=colors,
        min_value=min_value,
        max_value=max_value,
        logarithmic=logarithmic,
        color_map=color_map,
    )

    map_crs = cartopy.crs.PlateCarree()
    if map_axis is None:
        map_axis = pyplot.subplot(1, 1, 1, projection=map_crs)

    map_bounds = [
        float(nodes.coords['x'].min().values),
        float(nodes.coords['y'].min().values),
        float(nodes.coords['x'].max().values),
        float(nodes.coords['y'].max().values),
    ]

    countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))
    countries.plot(color='lightgrey', ax=map_axis)
    coast = cartopy.feature.NaturalEarthFeature(
        category='physical', scale='50m', facecolor='none', name='coastline'
    )
    map_axis.add_feature(coast, edgecolor='grey', linewidth=0.5)

    if storm is not None:
        if not isinstance(storm, VortexTrack):
            try:
                storm = VortexTrack.from_file(storm)
            except FileNotFoundError:
                storm = VortexTrack(storm)

        map_axis.plot(
            storm.data['longitude'], storm.data['latitude'], 'k--', label=storm.name,
        )

        if storm.name is not None:
            map_axis.legend(fontsize=6)

    if 'element' in nodes:
        mesh_tri = Triangulation(
            nodes.coords['x'],
            nodes.coords['y'],
            triangles=nodes.coords['element'],
            mask=numpy.isnan(nodes[data_var_name[0]][nodes.coords['element']]).any(axis=1),
        )
        levels = numpy.linspace(min_value, max_value, 26)
        map_axis.tricontourf(
            mesh_tri,
            nodes[data_var_name[0]].values,
            levels=levels,
            cmap=color_map,
            **kwargs,  # transform=map_crs,
        )
    else:
        map_axis.scatter(
            x=nodes.coords['x'],
            y=nodes.coords['y'],
            c=colors,
            s=2,
            norm=normalization,
            **kwargs,  # transform=map_crs,
        )

    map_axis.set_xlim(map_bounds[0], map_bounds[2])
    map_axis.set_ylim(map_bounds[1], map_bounds[3])

    if map_title is not None:
        map_axis.set_title(map_title)



# from matplotlib import pyplot
# from matplotlib.axis import Axis
# from matplotlib.cm import ScalarMappable
# from matplotlib.colors import Normalize
import numpy

# from ensembleperturbation.plotting.geometry import plot_points
# import gdal
# from affine import Affine


def bounds_from_opposite_corners(
    corner_1: (float, float), corner_2: (float, float)
) -> (float, float, float, float):
    """
    get bounds from two XY points

    :param corner_1: XY point
    :param corner_2: XY point
    :returns: min X, min Y, max X, max Y
    """

    return numpy.ravel(numpy.sort(numpy.stack((corner_1, corner_2), axis=0), axis=0))


def where_not_nodata(array: numpy.ndarray, nodata: float = None) -> numpy.ndarray:
    """
    get a boolean array of where data exists in the given array

    :param array: array of gridded data with dimensions (Z)YX
    :param nodata: value where there is no data in the given array
    :returns: array of booleans indicating where data exists
    """

    if nodata is None:
        if array.dtype == bool:
            nodata = False
        elif array.dtype == int:
            nodata = -2147483648
        else:
            nodata = numpy.nan

    coverage = array != nodata if not numpy.isnan(nodata) else ~numpy.isnan(array)

    if len(array.shape) > 2:
        coverage = numpy.any(coverage, axis=0)

    return coverage


def _maxValue(arr: numpy.ndarray):
    """
    returns the most used value in the array as an integer

    Takes an input array and finds the most used value in the array, this
    value is used by the program to assume the array's nodata value

    :param arr: An input array
    """

    nums, counts = numpy.unique(arr, return_counts=True)
    index = numpy.where(counts == numpy.amax(counts))
    return int(nums[index])


def geoarray_to_xyz(
    data: numpy.ndarray,
    origin: (float, float),
    resolution: (float, float),
    nodata: float = None,
) -> numpy.ndarray:
    """
    extract XYZ points from an array of data using the given raster-like georeference (origin  and resolution)

    :param data: 2D array of gridded data
    :param origin: X, Y coordinates of northwest corner
    :param resolution: cell size
    :param nodata: value to exclude from point creation from the input grid
    :returns: N x 3 array of XYZ points
    """

    if nodata is None:
        if data.dtype == float:
            nodata = numpy.nan
        elif data.dtype == int:
            nodata = -2147483648
        if data.dtype == bool:
            nodata = False

    data_coverage = where_not_nodata(data, nodata)
    x_values, y_values = numpy.meshgrid(
        numpy.linspace(origin[0], origin[0] + resolution[0] * data.shape[1], data.shape[1]),
        numpy.linspace(origin[1], origin[1] + resolution[1] * data.shape[0], data.shape[0]),
    )

    return numpy.stack(
        (x_values[data_coverage], y_values[data_coverage], data[data_coverage]), axis=1
    )


# def gdal_to_xyz(dataset: gdal.Dataset, nodata: float = None) -> numpy.ndarray:
#     """
#     extract XYZ points from a GDAL dataset
#
#     :param dataset: GDAL dataset (point cloud or raster)
#     :param nodata: value to exclude from point creation
#     :returns: N x M array of XYZ points
#     """
#
#     coordinates = None
#     layers_data = []
#     if dataset.RasterCount > 0:
#         for index in range(1, dataset.RasterCount + 1):
#             raster_band = dataset.GetRasterBand(index)
#             geotransform = dataset.GetGeoTransform()
#
#             if nodata is None:
#                 nodata = raster_band.GetNoDataValue()
#
#             points = geoarray_to_xyz(
#                 raster_band.ReadAsArray(),
#                 origin=(geotransform[0], geotransform[3]),
#                 resolution=(geotransform[1], geotransform[5]),
#                 nodata=nodata,
#             )
#             if coordinates is None:
#                 coordinates = points[:, :2]
#             layers_data.append(points[:, 2])
#     else:
#         for index in range(dataset.GetLayerCount()):
#             point_layer = dataset.GetLayerByIndex(index)
#             num_features = point_layer.GetFeatureCount()
#
#             for feature_index in range(num_features):
#                 feature = point_layer.GetFeature(feature_index)
#                 feature_geometry = feature.geometry()
#                 num_points = feature_geometry.GetGeometryCount()
#                 # TODO this assumes points in all layers are in the same order
#                 points = numpy.array(
#                     [
#                         feature_geometry.GetGeometryRef(point_index).GetPoint()
#                         for point_index in range(num_points)
#                         if feature_geometry.GetGeometryRef(point_index).GetPoint()[2] != nodata
#                     ]
#                 )
#
#                 if coordinates is None:
#                     coordinates = points[:, :2]
#                 layers_data.append(points[:, 2])
#
#     return numpy.concatenate(
#         [coordinates] + [numpy.expand_dims(data, axis=1) for data in layers_data], axis=1
#     )
#
#
# def gdal_raster_bounds(raster: gdal.Dataset) -> (float, float, float, float):
#     """
#     get the bounds (grouped by dimension) of the given unrotated raster
#
#     :param raster: GDAL raster dataset
#     :returns: min X, min Y, max X, max Y
#     """
#
#     geotransform = raster.GetGeoTransform()
#     origin = numpy.array((geotransform[0], geotransform[3]))
#     resolution = numpy.array((geotransform[1], geotransform[5]))
#     rotation = numpy.array((geotransform[2], geotransform[4]))
#     shape = raster.RasterYSize, raster.RasterXSize
#
#     if numpy.any(rotation != 0):
#         raise NotImplementedError('rotated rasters not supported')
#
#     return bounds_from_opposite_corners(origin, origin + numpy.flip(shape) * resolution)
#
#
# def plot_geoarray(
#     array: numpy.ndarray,
#     transform: Affine = None,
#     nodata: float = None,
#     axis: Axis = None,
#     show: bool = False,
#     **kwargs,
# ):
#     """
#     plot the given georeferenced array
#
#     :param array: 2D array of gridded data
#     :param transform: affine matrix transform
#     :param nodata: value representing no data in the given data
#     :param axis: `pyplot` axis to plot to
#     :param show: whether to show the plot
#     """
#
#     origin = (transform.c, transform.f)
#     resolution = (transform.a, transform.e)
#
#     if nodata is None:
#         nodata = numpy.nan
#
#     if axis is None:
#         axis = pyplot.gca()
#
#     if not numpy.isnan(nodata):
#         array[~where_not_nodata(array, nodata)] = numpy.nan
#
#     # if resolution[0] < 0:
#     #     data = numpy.flip(data, axis=1)
#     # if resolution[1] < 0:
#     #     data = numpy.flip(data, axis=0)
#
#     bounds = bounds_from_opposite_corners(
#         origin, origin + numpy.flip(array.shape) * resolution
#     )
#     axis.matshow(array, extent=bounds[[0, 2, 1, 3]], aspect='auto', **kwargs)
#
#     if show:
#         pyplot.show()
#
#
# def plot_dataset(
#     dataset: gdal.Dataset, index: int = 0, axis: Axis = None, show: bool = False, **kwargs,
# ):
#     """
#     plot the given GDAL dataset.
#
#     :param dataset: GDAL dataset (raster or point cloud)
#     :param index: zero-based index of raster band / vector layer to read
#     :param axis: `pyplot` axis to plot to
#     :param show: whether to show the plot
#     """
#
#     if dataset.RasterCount > 0:
#         transform = Affine.from_gdal(*dataset.GetGeoTransform())
#
#         raster_band = dataset.GetRasterBand(index + 1)
#         raster_data = raster_band.ReadAsArray()
#         nodata = raster_band.GetNoDataValue()
#         if nodata is None:
#             nodata = _maxValue(raster_data)
#         del raster_band
#
#         plot_geoarray(raster_data.astype('float64'), transform, nodata, axis, show, **kwargs)
#     else:
#         plot_points(gdal_to_xyz(dataset), index, axis, show, **kwargs)
#
#
# def plot_interpolation(
#     original_dataset: gdal.Dataset,
#     interpolated_raster: gdal.Dataset,
#     method: str,
#     input_index: int = 0,
#     output_index: int = 0,
#     show: bool = False,
# ):
#     """
#     plot original data side-by-side with an interpolated raster for comparison
#
#     :param original_dataset: GDAL dataset (point cloud or raster) of original data
#     :param interpolated_raster: GDAL raster of interpolated data
#     :param method: method of interpolation
#     :param input_index: zero-based index of layer / band to read from the input dataset
#     :param output_index: zero-based index of band to read from the output raster
#     :param show: whether to show the plot
#     """
#
#     if original_dataset.RasterCount > 0:
#         original_raster_band = original_dataset.GetRasterBand(input_index + 1)
#         original_data = original_raster_band.ReadAsArray()
#         original_nodata = original_raster_band.GetNoDataValue()
#         original_data[original_data == original_nodata] = numpy.nan
#     else:
#         original_data = gdal_to_xyz(original_dataset)[:, input_index + 2]
#
#     interpolated_raster_band = interpolated_raster.GetRasterBand(output_index + 1)
#     interpolated_nodata = interpolated_raster_band.GetNoDataValue()
#     interpolated_data = interpolated_raster_band.ReadAsArray()
#     interpolated_data[interpolated_data == interpolated_nodata] = numpy.nan
#
#     # get minimum and maximum values for all three dimensions
#     z_values = numpy.concatenate((numpy.ravel(interpolated_data), numpy.ravel(original_data)))
#     min_z = numpy.nanmin(z_values)
#     max_z = numpy.nanmax(z_values)
#
#     # create a new figure window with two subplots
#     figure = pyplot.figure()
#     left_axis = figure.add_subplot(1, 2, 1)
#     left_axis.set_title('survey data')
#     right_axis = figure.add_subplot(1, 2, 2, sharex=left_axis, sharey=left_axis)
#     right_axis.set_title(f'{method} interpolation to raster')
#
#     # plot data
#     plot_dataset(original_dataset, input_index, left_axis, vmin=min_z, vmax=max_z)
#     plot_dataset(interpolated_raster, input_index, right_axis, vmin=min_z, vmax=max_z)
#     right_axis.axes.get_yaxis().set_visible(False)
#
#     # create colorbar
#     figure.colorbar(
#         ScalarMappable(norm=Normalize(vmin=min_z, vmax=max_z)), ax=(right_axis, left_axis)
#     )
#
#     # pause program execution and show the figure
#     if show:
#         pyplot.show()



import math
from os import PathLike
from pathlib import Path
from typing import Dict, List

import cartopy
import geodatasets
import geopandas
from matplotlib import gridspec, pyplot
from matplotlib.axis import Axis
from matplotlib.colors import Normalize
from matplotlib.figure import Figure
import numpy
from stormevents.nhc import VortexTrack
import xarray

from ensembleperturbation.plotting.geometry import plot_points, plot_surface
from ensembleperturbation.plotting.nodes import plot_node_map
from ensembleperturbation.plotting.utilities import colorbar_axis


def get_validation_statistics(O, P, decimals):
    # pearson correlation coefficient
    def corrcoef(O, P):
        MP = P.mean()
        MO = O.mean()
        PD2 = ((P - MP) ** 2).sum()
        OD2 = ((O - MO) ** 2).sum()
        PDOD = ((P - MP) * (O - MO)).sum()
        corr = PDOD / (PD2 * OD2) ** 0.5
        if (PDOD == 0) & (PD2 == 0) & (OD2 == 0):
            corr.values = 1.0  # 0/0: perfect correlation
        elif (PD2 == 0) or (OD2 == 0):
            corr.values = 0.0  # No./0: no-correlation
        return corr

    # correlation coefficient
    corr = corrcoef(O, P)
    # wet-dry correlation coefficient
    OW = ~numpy.isnan(O)
    PW = ~numpy.isnan(P)
    corr_wd = corrcoef(OW, PW)
    # mean bias
    mb = (P - O).mean()
    # mean absolute error
    mae = (abs(P - O)).mean()
    # root-mean-square error
    rmse = (((P - O) ** 2).mean()) ** 0.5

    return (
        numpy.round(corr.values, decimals),
        numpy.round(mb.values, decimals),
        numpy.round(mae.values, decimals),
        numpy.round(rmse.values, decimals),
        numpy.round(corr_wd.values, decimals),
    )


def comparison_plot_grid(
    variables: List[str], figure: Figure = None
) -> (Dict[str, Dict[str, Axis]], gridspec.GridSpec):
    if figure is None:
        figure = pyplot.figure()

    num_variables = len(variables)
    num_plots = int(num_variables * (num_variables - 1) / 2)

    grid_length = math.ceil(num_plots ** 0.5)
    grid = gridspec.GridSpec(grid_length, grid_length, figure=figure, wspace=0, hspace=0)

    shared_grid_columns = {'x': {}, 'y': {}}
    for plot_index in range(num_plots):
        original_index = plot_index
        if plot_index >= num_variables:
            # TODO fix this
            plot_index = (plot_index % num_variables) + 1 if original_index % 2 else 3

        variable = variables[plot_index]

        if original_index % 2 == 0:
            shared_grid_columns['x'][variable] = None
        else:
            shared_grid_columns['y'][variable] = None

    if len(shared_grid_columns['y']) == 0:
        shared_grid_columns['y'][variables[-1]] = None

    axes = {}
    for row_index in range(grid_length):
        row_variable = list(shared_grid_columns['y'])[row_index]
        axes[row_variable] = {}
        for column_index in range(grid_length - row_index):
            column_variable = list(shared_grid_columns['x'])[column_index]

            sharex = shared_grid_columns['x'][column_variable]
            sharey = shared_grid_columns['y'][row_variable]

            variable_axis = figure.add_subplot(
                grid[row_index, column_index], sharex=sharex, sharey=sharey,
            )

            if sharex is None:
                shared_grid_columns['x'][column_variable] = variable_axis
            if sharey is None:
                shared_grid_columns['y'][row_variable] = variable_axis

            if grid_length != 1:
                if row_index == 0:
                    variable_axis.set_xlabel(column_variable)
                    variable_axis.xaxis.set_label_position('top')
                    variable_axis.xaxis.tick_top()
                    if row_index == grid_length - column_index - 1:
                        variable_axis.secondary_xaxis('bottom')
                elif row_index != grid_length - column_index - 1:
                    variable_axis.xaxis.set_visible(False)

                if column_index == 0:
                    variable_axis.set_ylabel(row_variable)
                    if row_index == grid_length - column_index - 1:
                        variable_axis.secondary_yaxis('right')
                elif row_index == grid_length - column_index - 1:
                    variable_axis.yaxis.tick_right()
                else:
                    variable_axis.yaxis.set_visible(False)
            else:
                variable_axis.set_xlabel(column_variable)
                variable_axis.set_ylabel(row_variable)

            axes[row_variable][column_variable] = variable_axis

    return axes, grid


def plot_scatter_comparison(
    nodes: xarray.DataArray,
    title: str = None,
    output_filename: PathLike = None,
    reference_line: bool = True,
    statistics_text_offset: int = 0,
    figure: Figure = None,
    axes: Dict[str, Dict[str, Axis]] = None,
    **kwargs,
):
    if 'source' not in nodes.dims:
        raise ValueError(f'"source" not found in data array dimensions: {nodes.dims}')
    elif len(nodes['source']) < 2:
        raise ValueError(f'cannot perform comparison with {len(nodes["source"])} source(s)')

    if 'c' not in kwargs and 'weights' in nodes:
        kwargs['c'] = nodes['weights']

    sources = nodes['source'].values

    if figure is None and axes is None:
        figure = pyplot.figure()
        figure.set_size_inches(10, 10 / 1.61803398875)

    if axes is None:
        axes, _ = comparison_plot_grid(sources, figure=figure)

    if title is not None:
        figure.suptitle(title)

    for row_source, columns in axes.items():
        for column_source, axis in columns.items():
            x = nodes.sel(source=column_source)
            y = nodes.sel(source=row_source)
            axis.scatter(x, y, **kwargs)

            if reference_line:
                min_value = numpy.nanmax([x.min(), y.min()])
                max_value = numpy.nanmin([x.max(), y.max()])
                axis.plot([min_value, max_value], [min_value, max_value], '--k', alpha=0.3)

            xlim = axis.get_xlim()
            ylim = axis.get_ylim()

            lim_min = min(xlim[0], ylim[0])
            lim_max = max(xlim[1], ylim[1])
            axis.set_xlim(lim_min, lim_max)
            axis.set_ylim(lim_min, lim_max)

            if statistics_text_offset > 0:
                ratio = statistics_text_offset * 0.1
                xlim = axis.get_xlim()
                ylim = axis.get_ylim()
                xpos = xlim[0] + ratio * (xlim[-1] - xlim[0])
                ypos = ylim[0] + numpy.array([0.95, 0.90, 0.85, 0.80]) * (ylim[-1] - ylim[0])
                corr, mb, mae, rmse, _ = get_validation_statistics(x, y, decimals=3)
                color = kwargs['c']
                axis.text(xpos, ypos[0], 'CORR = ' + str(corr), color=color)
                axis.text(xpos, ypos[1], 'MB = ' + str(mb) + ' m', color=color)
                axis.text(xpos, ypos[2], 'MAE = ' + str(mae) + ' m', color=color)
                axis.text(xpos, ypos[3], 'RMSE = ' + str(rmse) + ' m', color=color)

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')

    return figure, axes


def plot_boxplot_comparison(
    nodes: xarray.DataArray,
    title: str = None,
    output_filename: PathLike = None,
    subplot_integer: int = 111,
    figure: Figure = None,
    **kwargs,
):
    if 'source' not in nodes.dims:
        raise ValueError(f'"source" not found in data array dimensions: {nodes.dims}')
    elif len(nodes['source']) < 2:
        raise ValueError(f'cannot perform comparison with {len(nodes["source"])} source(s)')

    sources = nodes['source'].values
    dataset_type = nodes['type'].values

    if figure is None:
        figure = pyplot.figure()
        figure.set_size_inches(11, 11 / 1.61803398875)

    if title is not None:
        figure.suptitle(title)

    axis = figure.add_subplot(subplot_integer)

    num_runs = len(nodes['run'])
    rmse = numpy.empty(num_runs)
    corr = numpy.empty(num_runs)
    mb = numpy.empty(num_runs)
    mae = numpy.empty(num_runs)
    crwd = numpy.empty(num_runs)
    r = 0
    for run in nodes['run'].values:
        if not numpy.isnan(nodes.sel(run=run)).all():
            corr[r], mb[r], mae[r], rmse[r], crwd[r] = get_validation_statistics(
                nodes.sel(source='model', run=run),
                nodes.sel(source='surrogate', run=run),
                decimals=9,
            )
            r += 1

    c = 'gray'
    c2 = 'black'
    colors = ['blue', 'red']
    box = axis.boxplot(
        [corr[0:r], crwd[0:r], mb[0:r], mae[0:r], rmse[0:r]],
        whis=[5, 95],
        showfliers=True,
        patch_artist=True,
        widths=0.08,
        positions=[0.05, 0.2, 0.35, 0.5, 0.65],
        boxprops=dict(facecolor=c, color=c2, linewidth=0.25),
        whiskerprops=dict(color=c2, linewidth=1, linestyle='dashed'),
        medianprops=dict(color=c2, linewidth=1.5),
        capprops=dict(linewidth=1.5),
        flierprops=dict(marker='x', markerfacecolor=c2, markersize=2),
    )
    for cdx, cap in enumerate(box['caps']):
        cap.set(color=colors[numpy.mod(cdx, 2)])
    for median in box['medians']:
        data = median.get_data()
        axis.text(
            data[0][1], data[1][0], str(numpy.round(data[1][0], 2)), ha='left', va='center'
        )

    pyplot.grid(True)
    pyplot.ylim([-1, 1])
    pyplot.xlim([-0.01, 0.75])
    axis.title.set_text(f'{dataset_type}: {r} runs')
    axis.set_xticklabels(['CORR', 'CORR$_{w/d}$', 'MB [m]', 'MAE [m]', 'RMSE [m]'])

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')

    return figure


def plot_sensitivities(
    sensitivities: xarray.Dataset, storm: str = None, output_filename: PathLike = None
):
    figure = pyplot.figure()
    figure.set_size_inches(12, 12 / 1.61803398875)
    figure.suptitle(
        f'Sobol sensitivities of {len(sensitivities["variable"])} variable(s) and {len(sensitivities["order"])} order(s) along {len(sensitivities["node"])} node(s)'
    )

    grid = gridspec.GridSpec(
        len(sensitivities['order']),
        len(sensitivities['variable']),
        figure=figure,
        wspace=0,
        hspace=0,
    )
    map_crs = cartopy.crs.PlateCarree()

    for order_index, order in enumerate(sensitivities['order']):
        for variable_index, variable in enumerate(sensitivities['variable']):
            axis = figure.add_subplot(grid[order_index, variable_index], projection=map_crs)
            order_variable_sensitivities = sensitivities.sel(order=order, variable=variable)

            plot_node_map(
                order_variable_sensitivities,
                map_title=None,
                colors=order_variable_sensitivities['sensitivities']
                if 'element' not in sensitivities
                else None,
                storm=storm,
                map_axis=axis,
                min_value=0,
                max_value=1,
            )

            if variable_index == 0:
                axis.yaxis.set_visible(True)
                axis.set_ylabel(str(order.values))
            elif variable_index > 0:
                axis.yaxis.set_visible(False)

            if order_index == 0:
                axis.xaxis.set_visible(True)
                axis.set_xlabel(str(variable.values))
                axis.xaxis.set_label_position('top')
                axis.xaxis.tick_top()
            elif order_index > 0:
                axis.xaxis.set_visible(False)

    colorbar_axis(
        normalization=Normalize(vmin=0, vmax=1),
        axis=pyplot.axes([-1, 0.1, 3, 0.15]),
        orientation='horizontal',
        own_axis=True,
        color_map='plasma',
    )

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')


def plot_validations(validation: xarray.Dataset, output_directory: PathLike):
    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)

    validation = validation['results']
    sources = validation['source'].values

    type_colors = {'training': 'b', 'validation': 'r'}
    fig_scatter = None
    fig_boxplot = None
    axes = None
    ncols = min(len(validation['type']), 2)
    subplot_integer = math.ceil(len(validation['type']) / ncols) * 100 + ncols * 10 + 1
    for index, result_type in enumerate(validation['type'].values):
        result_validation = validation.sel(type=result_type)
        # fig_scatter, axes = plot_scatter_comparison(
        #    result_validation,
        #    title=f'comparison of {len(sources)} sources along {len(result_validation["node"])} node(s)',
        #    reference_line=index == 0,
        #    statistics_text_offset=2 * (index + 1),
        #    figure=fig_scatter,
        #    axes=axes,
        #    s=1,
        #    c=type_colors[result_type],
        #    label=result_type,
        # )

        fig_boxplot = plot_boxplot_comparison(
            result_validation,
            title=f'error statistics of surrogate model',
            subplot_integer=subplot_integer + index,
            figure=fig_boxplot,
        )

    # for row in axes.values():
    #    for axis in row.values():
    #        axis.legend()

    if output_directory is not None:
        # fig_scatter.savefig(
        #    output_directory / f'validation_scatter.png', dpi=200, bbox_inches='tight',
        # )
        fig_boxplot.savefig(
            output_directory / f'validation_boxplot.png', dpi=200, bbox_inches='tight',
        )


def plot_selected_validations(
    validation: xarray.Dataset, run_list: list, output_directory: PathLike
):
    sources = validation['source'].values
    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)

    bounds = numpy.array(
        [
            validation['x'].min(),
            validation['y'].min(),
            validation['x'].max(),
            validation['y'].max(),
        ]
    )
    vmax = numpy.round(validation.sel(source='model').results.quantile(0.98), decimals=1)
    vmin = 0.0
    for run in run_list:
        figure = pyplot.figure()
        figure.set_size_inches(10, 10 / 1.61803398875)
        figure.suptitle(f'validation of surrogate model for run: {run}')

        for index, source in enumerate(sources):
            map_axis = figure.add_subplot(2, len(sources), index + 1)
            map_axis.title.set_text(f'{source}')
            countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))

            map_axis.set_xlim((bounds[0], bounds[2]))
            map_axis.set_ylim((bounds[1], bounds[3]))

            xlim = map_axis.get_xlim()
            ylim = map_axis.get_ylim()

            countries.plot(color='lightgrey', ax=map_axis)

            points = numpy.vstack(
                (
                    validation['x'],
                    validation['y'],
                    validation.sel(type='validation', run=run, source=source).results,
                )
            ).T
            if 'element' not in validation:
                im = plot_points(
                    points=points, axis=map_axis, add_colorbar=False, vmax=vmax, vmin=vmin,
                )
            else:
                im = plot_surface(
                    points=points,
                    element_table=validation['element'].values,
                    axis=map_axis,
                    add_colorbar=False,
                    levels=numpy.linspace(vmin, vmax, 25 + 1),
                    extend='both',
                )

            map_axis.set_xlim(xlim)
            map_axis.set_ylim(ylim)

        pyplot.subplots_adjust(wspace=0.02, right=0.96)
        cax = pyplot.axes([0.95, 0.55, 0.015, 0.3])
        cbar = figure.colorbar(im, extend='both', cax=cax)
        cbar.ax.set_title('[m]')

        if output_directory is not None:
            figure.savefig(
                output_directory / f'validation_{run}.png', dpi=200, bbox_inches='tight',
            )


def plot_selected_percentiles(
    node_percentiles: xarray.Dataset,
    perc_list: list,
    output_directory: PathLike,
    storm: str = None,
):
    percentiles = node_percentiles.quantiles

    sources = node_percentiles['source'].values
    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)

    bounds = numpy.array(
        [
            node_percentiles['x'].min(),
            node_percentiles['y'].min(),
            node_percentiles['x'].max(),
            node_percentiles['y'].max(),
        ]
    )
    vmax = numpy.round(percentiles.sel(source='model').quantile(0.98), decimals=1)
    vmin = 0.0
    for perc in perc_list:
        figure = pyplot.figure()
        figure.set_size_inches(10, 10 / 1.61803398875)
        figure.suptitle(f'comparison of percentiles: {perc}%')
        for index, source in enumerate(sources):
            map_axis = figure.add_subplot(2, len(sources), index + 1)
            map_axis.title.set_text(f'{source}')
            countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))

            map_axis.set_xlim((bounds[0], bounds[2]))
            map_axis.set_ylim((bounds[1], bounds[3]))

            xlim = map_axis.get_xlim()
            ylim = map_axis.get_ylim()

            countries.plot(color='lightgrey', ax=map_axis)

            points = numpy.vstack(
                (
                    node_percentiles['x'],
                    node_percentiles['y'],
                    percentiles.sel(quantile=perc, source=source),
                )
            ).T
            if 'element' not in node_percentiles:
                im = plot_points(
                    points=points, axis=map_axis, add_colorbar=False, vmax=vmax, vmin=vmin,
                )
            else:
                im = plot_surface(
                    points=points,
                    element_table=node_percentiles['element'].values,
                    axis=map_axis,
                    add_colorbar=False,
                    levels=numpy.linspace(vmin, vmax, 25 + 1),
                    extend='both',
                )

            map_axis.set_xlim(xlim)
            map_axis.set_ylim(ylim)

            if storm is not None:
                if not isinstance(storm, VortexTrack):
                    try:
                        storm = VortexTrack.from_file(storm)
                    except FileNotFoundError:
                        storm = VortexTrack(storm)

                map_axis.plot(
                    storm.data['longitude'], storm.data['latitude'], 'k--', label=storm.name,
                )

                if storm.name is not None:
                    map_axis.legend(fontsize=6)

        pyplot.subplots_adjust(wspace=0.02, right=0.96)
        cax = pyplot.axes([0.95, 0.55, 0.015, 0.3])
        cbar = figure.colorbar(im, extend='both', cax=cax)
        cbar.ax.set_title('[m]')

        if output_directory is not None:
            figure.savefig(
                output_directory / f'percentiles_{perc}.png', dpi=200, bbox_inches='tight',
            )


def plot_kl_surrogate_fit(
    kl_fit: xarray.Dataset,
    output_filename: PathLike,
    reference_line: bool = True,
    statistics_text: bool = True,
):
    kl_fit = kl_fit['results']

    alim = [kl_fit.min(), kl_fit.max()]
    subplot_width = 3
    subplot_height = numpy.ceil(len(kl_fit['node']) / subplot_width).astype(int)

    figure = pyplot.figure()
    figure.set_size_inches(11, 11 * subplot_height / 5 / 1.61803398875)
    figure.suptitle(f'comparison of surrogate for the KL samples')

    for mode in range(len(kl_fit['node'])):
        axis = figure.add_subplot(subplot_height, subplot_width, mode + 1)
        qoi = kl_fit.sel(node=mode, source='model')
        qoi_pc = kl_fit.sel(node=mode, source='surrogate')

        axis.plot(qoi, qoi_pc, 'o', markersize=4)

        if reference_line:
            axis.plot([alim[0], alim[-1]], [alim[0], alim[-1]], '--k', alpha=0.3)

        if statistics_text:
            xpos = alim[0] + 0.1 * (alim[-1] - alim[0])
            ypos = alim[0].values + numpy.array([0.85, 0.7]) * (alim[-1] - alim[0]).values
            corr, _, _, rmse, _ = get_validation_statistics(qoi, qoi_pc, decimals=3)
            axis.text(xpos, ypos[0], 'RMSE = ' + str(rmse))
            axis.text(xpos, ypos[1], 'CORR = ' + str(corr))

        axis.set_xlim(alim)
        axis.set_ylim(alim)
        if mode + 1 > (subplot_height - 1) * subplot_width:
            axis.set_xlabel('actual')
        axis.set_ylabel('predicted')
        axis.title.set_text(f'KL mode-{mode + 1}')

    if output_filename is not None:
        figure.savefig(output_filename, dpi=200, bbox_inches='tight')


def plot_selected_probability_fields(
    node_prob_field: xarray.Dataset,
    level_list: list,
    output_directory: PathLike,
    label_unit_convert_factor: float = 1,
    label_unit_name: str = 'm',
    storm: str = None,
):
    probabilities = node_prob_field.probabilities

    sources = node_prob_field['source'].values
    if output_directory is not None:
        if not isinstance(output_directory, Path):
            output_directory = Path(output_directory)

    bounds = numpy.array(
        [
            node_prob_field['x'].min(),
            node_prob_field['y'].min(),
            node_prob_field['x'].max(),
            node_prob_field['y'].max(),
        ]
    )
    vmax = 1 + numpy.finfo(float).eps
    vmin = 0
    for lvl in level_list:
        figure = pyplot.figure()
        figure.set_size_inches(10, 10 / 1.61803398875)
        figure.suptitle(
            f'Probability of water level exceeding {round(lvl*label_unit_convert_factor)}-{label_unit_name}'
        )
        for index, source in enumerate(sources):
            map_axis = figure.add_subplot(2, len(sources), index + 1)
            map_axis.title.set_text(f'{source}')
            countries = geopandas.read_file(geodatasets.get_path('naturalearth land'))

            map_axis.set_xlim((bounds[0], bounds[2]))
            map_axis.set_ylim((bounds[1], bounds[3]))

            xlim = map_axis.get_xlim()
            ylim = map_axis.get_ylim()

            countries.plot(color='lightgrey', ax=map_axis)

            points = numpy.vstack(
                (
                    node_prob_field['x'],
                    node_prob_field['y'],
                    probabilities.sel(level=lvl, source=source),
                )
            ).T
            if 'element' not in node_prob_field:
                im = plot_points(
                    points=points, axis=map_axis, add_colorbar=False, vmax=vmax, vmin=vmin,
                )
            else:
                im = plot_surface(
                    points=points,
                    element_table=node_prob_field['element'].values,
                    axis=map_axis,
                    add_colorbar=False,
                    levels=numpy.linspace(vmin, vmax, 25),
                    extend='neither',
                )

            map_axis.set_xlim(xlim)
            map_axis.set_ylim(ylim)

            if storm is not None:
                if not isinstance(storm, VortexTrack):
                    try:
                        storm = VortexTrack.from_file(storm)
                    except FileNotFoundError:
                        storm = VortexTrack(storm)

                map_axis.plot(
                    storm.data['longitude'], storm.data['latitude'], 'k--', label=storm.name,
                )

                if storm.name is not None:
                    map_axis.legend(fontsize=6)

        pyplot.subplots_adjust(wspace=0.02, right=0.96)
        cax = pyplot.axes([0.95, 0.55, 0.015, 0.3])
        cbar = figure.colorbar(im, extend='neither', cax=cax)

        if output_directory is not None:
            figure.savefig(
                output_directory
                / f'probability_exceeding_{round(lvl*label_unit_convert_factor)}_{label_unit_name}.png',
                dpi=200,
                bbox_inches='tight',
            )



from os import PathLike
from typing import Iterable, List, Union

from matplotlib import pyplot, tri
from matplotlib.axis import Axis
from matplotlib.cm import get_cmap
import numpy
from shapely.geometry import MultiPoint, MultiPolygon, Polygon
from shapely.geometry import shape as shapely_shape


def plot_polygon(
    geometry: Union[Polygon, MultiPolygon], axis: Axis = None, show: bool = False, **kwargs,
):
    """
    plot the given polygon

    :param geometry: Shapely polygon (or multipolygon)
    :param axis: `pyplot` axis to plot to
    :param show: whether to show the plot
    """

    if axis is None:
        axis = pyplot.gca()

    if 'c' not in kwargs:
        try:
            color = next(axis._get_lines.color_cycle)
        except AttributeError:
            color = 'r'
        kwargs['c'] = color

    if isinstance(geometry, dict):
        geometry = shapely_shape(geometry)

    if isinstance(geometry, Polygon):
        axis.plot(*geometry.exterior.xy, **kwargs)
        for interior in geometry.interiors:
            axis.plot(*interior.xy, **kwargs)
    elif isinstance(geometry, MultiPolygon):
        for polygon in geometry:
            plot_polygon(polygon, axis, show=False, **kwargs)
    else:
        axis.plot(*geometry.xy, **kwargs)

    if show:
        pyplot.show()


def plot_polygons(
    geometries: List[Polygon],
    colors: List[str] = None,
    axis: Axis = None,
    show: bool = False,
    **kwargs,
):
    """
    plot the given polygons using the given colors

    :param geometries: list of shapely polygons or multipolygons
    :param colors: colors to plot each region
    :param axis: `pyplot` axis to plot to
    :param show: whether to show the plot
    """

    if axis is None:
        axis = pyplot.gca()

    if 'c' in kwargs:
        colors = [kwargs['c'] for _ in range(len(geometries))]
    elif colors is None:
        colors = [
            get_cmap('gist_rainbow')(color_index / len(geometries))
            for color_index in range(len(geometries))
        ]

    for geometry_index, geometry in enumerate(geometries):
        kwargs['c'] = colors[geometry_index]
        plot_polygon(geometry, axis, **kwargs)

    if show:
        pyplot.show()


def plot_bounding_box(
    sw: (float, float), ne: (float, float), axis: Axis = None, show: bool = False, **kwargs,
):
    """
    plot the bounding box of the given extent

    :param sw: XY coordinates of southwest corner
    :param ne: XY coordinates of northeast corner
    :param axis: `pyplot` axis to plot to
    :param show: whether to show the plot
    """

    if axis is None:
        axis = pyplot.gca()

    corner_points = numpy.array([sw, (ne[0], sw[1]), ne, (sw[0], ne[1]), sw])

    axis.plot(corner_points[:, 0], corner_points[:, 1], **kwargs)

    if show:
        pyplot.show()


def plot_points(
    points: Union[numpy.ndarray, MultiPoint],
    index: int = 0,
    axis: Axis = None,
    show: bool = False,
    save_filename: PathLike = None,
    title: str = None,
    add_colorbar: bool = True,
    **kwargs,
):
    """
    create a scatter plot of the given points

    :param points: N x M array of points
    :param index: zero-based index of vector layer to read
    :param axis: `pyplot` axis to plot to
    :param show: whether to show the plot
    :param save_filename: whether to save the plot
    :param title: whether to add a title to the plot
    """

    if isinstance(points, MultiPoint):
        points = numpy.squeeze(
            numpy.stack([list(point.coords)[0][:3] for point in points.geoms], axis=0)
        )

    if not isinstance(points, numpy.ndarray):
        points = numpy.array(points)

    if axis is None:
        axis = pyplot.gca()

    if 'c' not in kwargs and points.shape[1] > 2:
        kwargs['c'] = points[:, index + 2]

    if 's' not in kwargs:
        kwargs['s'] = 2

    cb_kwarg = {}
    if 'extend' in kwargs:
        cb_kwarg = {'extend': kwargs['extend']}
        kwargs.pop('extend')

    sc = axis.scatter(points[:, 0], points[:, 1], **kwargs)

    if 'c' in kwargs and add_colorbar:
        pyplot.colorbar(sc, shrink=0.8, **cb_kwarg)

    if title is not None:
        pyplot.title(title)

    if save_filename is not None:
        pyplot.savefig(save_filename)

    if show:
        pyplot.show()

    return sc


def plot_surface(
    points: numpy.ndarray,
    element_table: numpy.ndarray,
    depth: numpy.ndarray = None,
    axis: Axis = None,
    show: bool = False,
    save_filename: PathLike = None,
    title: str = None,
    add_colorbar: bool = True,
    **kwargs,
):
    """
    create a trisurface plot of the given points

    :param points: N x 3 array of coordinates and heights
    :param element_table: NE x 3 array for element table
    :param axis: `pyplot` axis to plot to
    :param show: whether to show the plot
    :param save_filename: whether to save the plot
    :param title: whether to add a title to the plot
    """

    mesh_tri = tri.Triangulation(
        points[:, 0],
        points[:, 1],
        triangles=element_table,
        mask=numpy.isnan(points[:, 2][element_table]).any(axis=1),
    )

    if axis is None:
        axis = pyplot.gca()

    tc = axis.tricontourf(mesh_tri, points[:, 2], **kwargs)

    if add_colorbar:
        pyplot.colorbar(tc, shrink=0.8)

    if title is not None:
        pyplot.title(title)

    if save_filename is not None:
        pyplot.savefig(save_filename)

    if show:
        pyplot.show()

    return tc
